"""
Mirror of the gene tests but two main differences:
> asynchronous
> implemented in pytest for asyncio marker
"""

import logging
import types

import pytest


import biothings_client
from biothings_client.client.definitions import AsyncMyGeneInfo
from biothings_client.utils.score import descore


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@pytest.mark.asyncio
async def test_metadata(async_gene_client: AsyncMyGeneInfo):
    gene_metadata = await async_gene_client.metadata()
    assert "stats" in gene_metadata
    assert "total_genes" in gene_metadata["stats"]


@pytest.mark.asyncio
async def test_getgene(async_gene_client: AsyncMyGeneInfo):
    g = await async_gene_client.getgene("1017")
    assert g["_id"] == "1017"
    assert g["symbol"] == "CDK2"


@pytest.mark.asyncio
async def test_getgene_with_fields(async_gene_client: AsyncMyGeneInfo):
    g = await async_gene_client.getgene("1017", fields="name,symbol,refseq")
    assert "_id" in g
    assert "name" in g
    assert "symbol" in g
    assert "refseq" in g
    assert "summary" not in g


@pytest.mark.asyncio
async def test_curie_id_query(async_gene_client: AsyncMyGeneInfo):
    """
    Tests the annotation endpoint support for the biolink CURIE ID.

    If support is enabled then we should retrieve the exact same document for all the provided
    queries
    """
    curie_id_testing_collection = [
        ("1017", "entrezgene:1017", "NCBIgene:1017"),
        (1017, "entrezgene:1017", "ncbigene:1017"),
        ("1017", "entrezgene:1017", "NCBIGENE:1017"),
        ("1018", "ensembl.gene:ENSG00000250506", "ENSEMBL:ENSG00000250506"),
        (1018, "ensembl.gene:ENSG00000250506", "ensembl:ENSG00000250506"),
        ("5995", "uniprot.Swiss-Prot:P47804", "UniProtKB:P47804"),
        (5995, "uniprot.Swiss-Prot:P47804", "UNIPROTKB:P47804"),
        ("5995", "uniprot.Swiss-Prot:P47804", "uniprotkb:P47804"),
    ]

    results_aggregation = []
    for id_query, biothings_query, biolink_query in curie_id_testing_collection:
        id_query_result = await async_gene_client.getgene(_id=id_query)
        biothings_term_query_result = await async_gene_client.getgene(_id=biothings_query)
        biolink_term_query_result = await async_gene_client.getgene(_id=biolink_query)
        results_aggregation.append(
            (
                id_query_result == biothings_term_query_result,
                id_query_result == biolink_term_query_result,
                biothings_term_query_result == biolink_term_query_result,
            )
        )

    results_validation = []
    failure_messages = []
    for result, test_query in zip(results_aggregation, curie_id_testing_collection):
        cumulative_result = all(result)
        if not cumulative_result:
            failure_messages.append(f"Query Failure: {test_query} | Results: {result}")
        results_validation.append(cumulative_result)

    assert all(results_validation), "\n".join(failure_messages)


@pytest.mark.asyncio
async def test_multiple_curie_id_query(async_gene_client: AsyncMyGeneInfo):
    """
    Tests the annotations endpoint support for the biolink CURIE ID.

    Batch query testing against the POST endpoint to verify that the CURIE ID can work with
    multiple

    If support is enabled then we should retrieve the exact same document for all the provided
    queries
    """
    curie_id_testing_collection = [
        ("1017", "entrezgene:1017", "NCBIgene:1017"),
        (1017, "entrezgene:1017", "ncbigene:1017"),
        ("1017", "entrezgene:1017", "NCBIGENE:1017"),
        ("1018", "ensembl.gene:ENSG00000250506", "ENSEMBL:ENSG00000250506"),
        (1018, "ensembl.gene:ENSG00000250506", "ensembl:ENSG00000250506"),
        ("5995", "uniprot.Swiss-Prot:P47804", "UniProtKB:P47804"),
        (5995, "uniprot.Swiss-Prot:P47804", "UNIPROTKB:P47804"),
        ("5995", "uniprot.Swiss-Prot:P47804", "uniprotkb:P47804"),
    ]

    results_aggregation = []
    for id_query, biothings_query, biolink_query in curie_id_testing_collection:
        base_result = await async_gene_client.getgene(_id=id_query)

        batch_query = [id_query, biothings_query, biolink_query]
        query_results = await async_gene_client.getgenes(ids=batch_query)
        assert len(query_results) == len(batch_query)

        batch_id_query = query_results[0]
        batch_biothings_query = query_results[1]
        batch_biolink_query = query_results[2]

        batch_id_query_return_value = batch_id_query.pop("query")
        assert batch_id_query_return_value == str(id_query)

        batch_biothings_query_return_value = batch_biothings_query.pop("query")
        assert batch_biothings_query_return_value == str(biothings_query)

        batch_biolink_query_return_value = batch_biolink_query.pop("query")
        assert batch_biolink_query_return_value == str(biolink_query)

        batch_result = (
            base_result == batch_id_query,
            base_result == batch_biothings_query,
            base_result == batch_biolink_query,
        )
        results_aggregation.append(batch_result)

    results_validation = []
    failure_messages = []
    for result, test_query in zip(results_aggregation, curie_id_testing_collection):
        cumulative_result = all(result)
        if not cumulative_result:
            failure_messages.append(f"Query Failure: {test_query} | Results: {result}")
        results_validation.append(cumulative_result)

    assert all(results_validation), "\n".join(failure_messages)


@pytest.mark.asyncio
async def test_getgene_with_fields_as_list(async_gene_client: AsyncMyGeneInfo):
    g1 = await async_gene_client.getgene("1017", fields="name,symbol,refseq")
    g2 = await async_gene_client.getgene("1017", fields=["name", "symbol", "refseq"])
    assert g1 == g2


@pytest.mark.asyncio
async def test_getgenes(async_gene_client: AsyncMyGeneInfo):
    g_li = await async_gene_client.getgenes([1017, 1018, "ENSG00000148795"])
    assert len(g_li) == 3
    assert g_li[0]["_id"] == "1017"
    assert g_li[1]["_id"] == "1018"
    assert g_li[2]["_id"] == "1586"


@pytest.mark.asyncio
async def test_query(async_gene_client: AsyncMyGeneInfo):
    qres = await async_gene_client.query("cdk2", size=5)
    assert "hits" in qres
    assert len(qres["hits"]) == 5


@pytest.mark.asyncio
async def test_query_with_fields_as_list(async_gene_client: AsyncMyGeneInfo):
    qres1 = await async_gene_client.query("entrezgene:1017", fields="name,symbol,refseq")
    qres2 = await async_gene_client.query("entrezgene:1017", fields=["name", "symbol", "refseq"])
    assert "hits" in qres1
    assert len(qres1["hits"]) == 1
    assert descore(qres1["hits"]) == descore(qres2["hits"])


@pytest.mark.asyncio
async def test_query_reporter(async_gene_client: AsyncMyGeneInfo):
    qres = await async_gene_client.query("reporter:1000_at")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "5595"


@pytest.mark.asyncio
async def test_query_symbol(async_gene_client: AsyncMyGeneInfo):
    qres = await async_gene_client.query("symbol:cdk2", species="mouse")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "12566"


@pytest.mark.asyncio
async def test_query_fetch_all(async_gene_client: AsyncMyGeneInfo):
    qres = await async_gene_client.query("_exists_:pdb")
    total = qres["total"]

    qres_generator = await async_gene_client.query("_exists_:pdb", fields="pdb", fetch_all=True)
    assert isinstance(qres_generator, types.AsyncGeneratorType)

    async_count = 0
    async for async_res in qres_generator:
        async_count += 1
    assert total == async_count


@pytest.mark.asyncio
async def test_querymany(async_gene_client: AsyncMyGeneInfo):
    qres = await async_gene_client.querymany([1017, "695"], verbose=False)
    assert len(qres) == 2

    qres = await async_gene_client.querymany("1017,695", verbose=False)
    assert len(qres) == 2


@pytest.mark.asyncio
async def test_querymany_with_scopes(async_gene_client: AsyncMyGeneInfo):
    qres = await async_gene_client.querymany([1017, "695"], scopes="entrezgene", verbose=False)
    assert len(qres) == 2

    qres = await async_gene_client.querymany([1017, "BTK"], scopes="entrezgene,symbol", verbose=False)
    assert len(qres) >= 2


@pytest.mark.asyncio
async def test_querymany_species(async_gene_client: AsyncMyGeneInfo):
    qres = await async_gene_client.querymany([1017, "695"], scopes="entrezgene", species="human", verbose=False)
    assert len(qres) == 2

    qres = await async_gene_client.findgenes([1017, "695"], scopes="entrezgene", species=9606, verbose=False)
    assert len(qres) == 2

    qres = await async_gene_client.findgenes([1017, "CDK2"], scopes="entrezgene,symbol", species=9606, verbose=False)
    assert len(qres) == 2


@pytest.mark.asyncio
async def test_querymany_fields(async_gene_client: AsyncMyGeneInfo):
    qres1 = await async_gene_client.findgenes(
        [1017, "CDK2"], scopes="entrezgene,symbol", fields=["uniprot", "unigene"], species=9606, verbose=False
    )
    assert len(qres1) == 2

    qres2 = await async_gene_client.findgenes(
        "1017,CDK2", scopes="entrezgene,symbol".split(","), fields="uniprot,unigene", species=9606, verbose=False
    )
    assert len(qres2) == 2
    assert descore(qres1) == descore(qres2)


@pytest.mark.asyncio
async def test_querymany_notfound(async_gene_client: AsyncMyGeneInfo):
    qres = await async_gene_client.findgenes([1017, "695", "NA_TEST"], scopes="entrezgene", species=9606)
    assert len(qres) == 3
    assert qres[2] == {"query": "NA_TEST", "notfound": True}


@pytest.mark.asyncio
@pytest.mark.skipif(not biothings_client._PANDAS, reason="requires the pandas library")
async def test_querymany_dataframe(async_gene_client: AsyncMyGeneInfo):
    from pandas import DataFrame

    query_list = [
        "1007_s_at",
        "1053_at",
        "117_at",
        "121_at",
        "1255_g_at",
        "1294_at",
        "1316_at",
        "1320_at",
        "1405_i_at",
        "1431_at",
    ]

    qres = await async_gene_client.querymany(query_list, scopes="reporter", as_dataframe=True)
    assert isinstance(qres, DataFrame)
    assert "name" in qres.columns
    assert set(query_list), set(qres.index)


@pytest.mark.asyncio
async def test_querymany_step(async_gene_client: AsyncMyGeneInfo):
    query_list = [
        "1007_s_at",
        "1053_at",
        "117_at",
        "121_at",
        "1255_g_at",
        "1294_at",
        "1316_at",
        "1320_at",
        "1405_i_at",
        "1431_at",
    ]
    qres1 = await async_gene_client.querymany(query_list, scopes="reporter")
    default_step = async_gene_client.step
    async_gene_client.step = 3
    qres2 = await async_gene_client.querymany(query_list, scopes="reporter")
    async_gene_client.step = default_step
    qres1 = descore(sorted(qres1, key=lambda doc: doc["_id"]))
    qres2 = descore(sorted(qres2, key=lambda doc: doc["_id"]))
    assert qres1 == qres2


@pytest.mark.asyncio
async def test_get_fields(async_gene_client: AsyncMyGeneInfo):
    fields = await async_gene_client.get_fields()
    assert "uniprot" in fields.keys()
    assert "exons" in fields.keys()

    fields = await async_gene_client.get_fields("kegg")
    assert "pathway.kegg" in fields.keys()



"""
Package versioning using PEP-376
https://peps.python.org/pep-0396/#pep-376-metadata
"""

try:
    from importlib import metadata
except ImportError:
    # Running on pre-3.8 Python; use importlib-metadata package
    import importlib_metadata as metadata  # type: ignore

DISTRIBUTION_NAME = "biothings_client"
__version__ = metadata.version(DISTRIBUTION_NAME)



"""
Tests for exercising the sychronous biothings_client for mygene
"""

import logging
import types

import pytest

import biothings_client
from biothings_client.utils.score import descore
from biothings_client.client.definitions import MyVariantInfo

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def test_format_hgvs(variant_client: MyVariantInfo):
    assert variant_client.format_hgvs("1", 35366, "C", "T") == "chr1:g.35366C>T"
    assert variant_client.format_hgvs("chr2", 17142, "G", "GA") == "chr2:g.17142_17143insA"
    assert variant_client.format_hgvs("1", 10019, "TA", "T") == "chr1:g.10020del"
    assert variant_client.format_hgvs("MT", 8270, "CACCCCCTCT", "C") == "chrMT:g.8271_8279del"
    assert variant_client.format_hgvs("7", 15903, "G", "GC") == "chr7:g.15903_15904insC"
    assert variant_client.format_hgvs("X", 107930849, "GGA", "C") == "chrX:g.107930849_107930851delinsC"
    assert variant_client.format_hgvs("20", 1234567, "GTC", "GTCT") == "chr20:g.1234569_1234570insT"


def test_metadata(variant_client: MyVariantInfo):
    meta = variant_client.metadata()
    assert "stats" in meta
    assert "total" in meta["stats"]


def test_getvariant(variant_client: MyVariantInfo):
    v = variant_client.getvariant("chr9:g.107620835G>A")
    assert v["_id"] == "chr9:g.107620835G>A"
    assert v["snpeff"]["ann"]["genename"] == "ABCA1"

    v = variant_client.getvariant("'chr1:g.1A>C'")  # something does not exist
    assert v is None


def test_getvariant_with_fields(variant_client: MyVariantInfo):
    v = variant_client.getvariant("chr9:g.107620835G>A", fields="dbnsfp,cadd,cosmic")
    assert "_id" in v
    assert "dbnsfp" in v
    assert "cadd" in v
    assert "cosmic" in v


def test_getvariants(variant_client: MyVariantInfo):
    query_list1 = [
        "chr1:g.866422C>T",
        "chr1:g.876664G>A",
        "chr1:g.69635G>C",
        "chr1:g.69869T>A",
        "chr1:g.881918G>A",
        "chr1:g.865625G>A",
        "chr1:g.69892T>C",
        "chr1:g.879381C>T",
        "chr1:g.878330C>G",
    ]
    v_li = variant_client.getvariants(query_list1)
    assert len(v_li) == 9
    assert v_li[0]["_id"] == query_list1[0]
    assert v_li[1]["_id"] == query_list1[1]
    assert v_li[2]["_id"] == query_list1[2]

    variant_client.step = 4
    # test input is a string of comma-separated ids
    v_li2 = variant_client.getvariants(",".join(query_list1))
    assert v_li == v_li2
    # test input is a tuple
    v_li2 = variant_client.getvariants(tuple(query_list1))
    assert v_li == v_li2

    # test input is a generator
    def _input(li):
        for x in li:
            yield x

    v_li2 = variant_client.getvariants(_input(query_list1))
    assert v_li == v_li2
    variant_client.step = 1000


def test_query(variant_client: MyVariantInfo):
    qres = variant_client.query("dbnsfp.genename:cdk2", size=5)
    assert "hits" in qres
    assert len(qres["hits"]) == 5


def test_query_hgvs(variant_client: MyVariantInfo):
    qres = variant_client.query('"NM_000048.4:c.566A>G"', size=5)
    # should match clinvar.hgvs.coding field from variant "chr7:g.65551772A>G"
    # sometime we need to update ".4" part if clinvar data updated.
    assert "hits" in qres
    assert len(qres["hits"]) == 1


def test_query_rsid(variant_client: MyVariantInfo):
    qres = variant_client.query("dbsnp.rsid:rs58991260")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "chr1:g.218631822G>A"
    qres2 = variant_client.query("rs58991260")

    # exclude _score field before comparison
    qres["hits"][0].pop("_score")
    qres2["hits"][0].pop("_score")
    assert qres["hits"] == qres2["hits"]


def test_query_symbol(variant_client: MyVariantInfo):
    qres = variant_client.query("snpeff.ann.genename:cdk2")
    assert "hits" in qres
    assert qres["total"] > 5000
    assert qres["hits"][0]["snpeff"]["ann"][0]["genename"] == "CDK2"


def test_query_genomic_range(variant_client: MyVariantInfo):
    qres = variant_client.query("chr1:69000-70000")
    assert "hits" in qres
    assert qres["total"] >= 3


def test_query_fetch_all(variant_client: MyVariantInfo):
    qres = variant_client.query("chr1:69500-70000", fields="chrom")
    total = qres["total"]

    qres = variant_client.query("chr1:69500-70000", fields="chrom", fetch_all=True)
    assert isinstance(qres, types.GeneratorType)
    assert total == len(list(qres))


def test_querymany(variant_client: MyVariantInfo):
    query_list1 = [
        "chr1:g.866422C>T",
        "chr1:g.876664G>A",
        "chr1:g.69635G>C",
        "chr1:g.69869T>A",
        "chr1:g.881918G>A",
        "chr1:g.865625G>A",
        "chr1:g.69892T>C",
        "chr1:g.879381C>T",
        "chr1:g.878330C>G",
    ]
    qres = variant_client.querymany(query_list1, verbose=False)
    assert len(qres) == 9

    variant_client.step = 4

    # test input as a string
    qres2 = variant_client.querymany(",".join(query_list1), verbose=False)
    assert qres == qres2

    # test input as a tuple
    qres2 = variant_client.querymany(tuple(query_list1), verbose=False)
    assert qres == qres2

    # test input as a iterator
    qres2 = variant_client.querymany(iter(query_list1), verbose=False)
    assert qres == qres2
    variant_client.step = 1000


def test_querymany_with_scopes(variant_client: MyVariantInfo):
    qres = variant_client.querymany(["rs58991260", "rs2500"], scopes="dbsnp.rsid", verbose=False)
    assert len(qres) == 2

    qres = variant_client.querymany(
        ["RCV000083620", "RCV000083611", "RCV000083584"], scopes="clinvar.rcv_accession", verbose=False
    )
    assert len(qres) == 3

    qres = variant_client.querymany(
        ["rs2500", "RCV000083611", "COSM1392449"],
        scopes="clinvar.rcv_accession,dbsnp.rsid,cosmic.cosmic_id",
        verbose=False,
    )
    assert len(qres) == 3


def test_querymany_fields(variant_client: MyVariantInfo):
    ids = ["COSM1362966", "COSM990046", "COSM1392449"]
    qres1 = variant_client.querymany(
        ids, scopes="cosmic.cosmic_id", fields=["cosmic.tumor_site", "cosmic.cosmic_id"], verbose=False
    )
    assert len(qres1) == 3

    qres2 = variant_client.querymany(
        ids, scopes="cosmic.cosmic_id", fields="cosmic.tumor_site,cosmic.cosmic_id", verbose=False
    )
    assert len(qres2) == 3

    assert descore(qres1) == descore(qres2)


def test_querymany_notfound(variant_client: MyVariantInfo):
    qres = variant_client.querymany(["rs58991260", "rs2500", "NA_TEST"], scopes="dbsnp.rsid", verbose=False)
    assert len(qres) == 3
    assert qres[2] == {"query": "NA_TEST", "notfound": True}


@pytest.mark.skipif(not biothings_client._PANDAS, reason="pandas library not installed")
def test_querymany_dataframe(variant_client: MyVariantInfo):
    from pandas import DataFrame

    query_list2 = [
        "rs374802787",
        "rs1433078",
        "rs1433115",
        "rs377266517",
        "rs587640013",
        "rs137857980",
        "rs199710579",
        "rs186823979",
        # 'rs2276240',
        "rs34521797",
        "rs372452565",
    ]

    qres = variant_client.querymany(query_list2, scopes="dbsnp.rsid", fields="dbsnp", as_dataframe=True, verbose=False)
    assert isinstance(qres, DataFrame)
    assert "dbsnp.vartype" in qres.columns
    assert set(query_list2) == set(qres.index)


def test_querymany_step(variant_client: MyVariantInfo):
    query_list2 = [
        "rs374802787",
        "rs1433078",
        "rs1433115",
        "rs377266517",
        "rs587640013",
        "rs137857980",
        "rs199710579",
        "rs186823979",
        # 'rs2276240',
        "rs34521797",
        "rs372452565",
    ]
    qres1 = variant_client.querymany(query_list2, scopes="dbsnp.rsid", fields="dbsnp.rsid", verbose=False)
    default_step = variant_client.step
    variant_client.step = 3
    qres2 = variant_client.querymany(query_list2, scopes="dbsnp.rsid", fields="dbsnp.rsid", verbose=False)
    variant_client.step = default_step
    # assert qres1, qres2, (qres1, qres2))
    assert descore(qres1) == descore(qres2)


def test_get_fields(variant_client: MyVariantInfo):
    fields = variant_client.get_fields()
    assert "dbsnp.chrom" in fields.keys()
    assert "clinvar.chrom" in fields.keys()



import sys
import importlib


_PANDAS = importlib.util.find_spec("pandas") is not None
_CACHING = importlib.util.find_spec("hishel") is not None and importlib.util.find_spec("anysqlite") is not None
_CACHING_NOT_SUPPORTED = sys.version_info < (3, 8)



"""
Tests for exercising the sychronous biothings_client for mygeneset
"""

import logging
import types

import pytest

import biothings_client
from biothings_client.utils.score import descore
from biothings_client.client.definitions import MyGenesetInfo


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def test_metadata(geneset_client: MyGenesetInfo):
    meta = geneset_client.metadata()
    assert "src" in meta
    assert "stats" in meta
    assert "total" in meta["stats"]


def test_getgeneset(geneset_client: MyGenesetInfo):
    gs = geneset_client.getgeneset("WP100")
    assert gs["_id"] == "WP100"
    assert gs["name"] == "Glutathione metabolism"
    assert gs["source"] == "wikipathways"
    assert gs["taxid"] == "9606"
    assert len(gs["genes"]) >= 19
    assert gs["count"] == len(gs["genes"])

    assert "wikipathways" in gs
    assert gs["wikipathways"]["id"] == "WP100"
    assert gs["wikipathways"]["pathway_name"] == "Glutathione metabolism"
    assert gs["wikipathways"]["url"] == "https://www.wikipathways.org/instance/WP100"
    assert gs["wikipathways"]["_license"] == "https://www.wikipathways.org/terms.html"

    assert any((gene.get("mygene_id") == "2937" and gene.get("symbol") == "GSS") for gene in gs["genes"])


def test_query_fetch_all(geneset_client: MyGenesetInfo):
    """
    pdb --> reactome
    q = source:reactome
    _exists_:pdb ---> source:reactome
    """
    qres = geneset_client.query("source:reactome")
    total = qres["total"]

    qres = geneset_client.query("source:reactome", fields="source,count,name", fetch_all=True)
    assert isinstance(qres, types.GeneratorType)
    assert total == len(list(qres))


def test_query_with_fields_as_list(geneset_client: MyGenesetInfo):
    qres1 = geneset_client.query("genes.ncbigene:1017", fields="name,source,taxid")
    qres2 = geneset_client.query("genes.ncbigene:1017", fields=["name", "source", "taxid"])
    assert "hits" in qres1
    assert len(qres1["hits"]) == 10
    assert descore(qres1["hits"]) == descore(qres2["hits"])


def test_getgeneset_with_fields(geneset_client: MyGenesetInfo):
    gs = geneset_client.getgeneset("WP100", fields="name,source,taxid,genes.mygene_id,genes.symbol")

    assert "_id" in gs
    assert "name" in gs
    assert "source" in gs
    assert "taxid" in gs

    assert any((gene.get("mygene_id") == "2937" and gene.get("symbol") == "GSS") for gene in gs["genes"])
    assert not any(gene.get("name") for gene in gs["genes"])


def test_getgenesets(geneset_client: MyGenesetInfo):
    gs_li = geneset_client.getgenesets(["WP100", "WP101", "WP103"])

    assert len(gs_li) == 3
    assert gs_li[0]["_id"] == "WP100"
    assert gs_li[1]["_id"] == "WP101"
    assert gs_li[2]["_id"] == "WP103"


def test_query(geneset_client: MyGenesetInfo):
    qres = geneset_client.query("genes.mygene_id:2937", size=5)
    assert "hits" in qres
    assert len(qres["hits"]) == 5


def test_query_default_fields(geneset_client: MyGenesetInfo):
    geneset_client.query(q="glucose")


def test_query_field(geneset_client: MyGenesetInfo):
    geneset_client.query(q="genes.ncbigene:1017")


def test_species_filter_plus_query(geneset_client: MyGenesetInfo):
    dog = geneset_client.query(q="glucose", species="9615")
    assert dog["hits"][0]["taxid"] == "9615"


def test_query_by_id(geneset_client: MyGenesetInfo):
    query = geneset_client.query(q="_id:WP100")
    assert query["hits"][0]["_id"] == "WP100"


def test_query_by_name(geneset_client: MyGenesetInfo):
    kinase = geneset_client.query(q="name:kinase")
    assert "kinase" in kinase["hits"][0]["name"].lower()


def test_query_by_description(geneset_client: MyGenesetInfo):
    desc = geneset_client.query(q="description:cytosine deamination")
    assert "cytosine" in desc["hits"][0]["description"].lower()
    assert "deamination" in desc["hits"][0]["description"].lower()


def test_query_by_source_go(geneset_client: MyGenesetInfo):
    go = geneset_client.query(q="source:go", fields="all")
    assert "go" in go["hits"][0].keys()
    assert go["hits"][0]["source"] == "go"


def test_query_by_source_ctd(geneset_client: MyGenesetInfo):
    ctd = geneset_client.query(q="source:ctd", fields="all")
    assert "ctd" in ctd["hits"][0].keys()
    assert ctd["hits"][0]["source"] == "ctd"


def test_query_by_source_msigdb(geneset_client: MyGenesetInfo):
    msigdb = geneset_client.query(q="source:msigdb", fields="all")
    assert "msigdb" in msigdb["hits"][0].keys()
    assert msigdb["hits"][0]["source"] == "msigdb"


@pytest.mark.xfail(reason="We removed kegg data source for now")
def test_query_by_source_kegg(geneset_client: MyGenesetInfo):
    kegg = geneset_client.query(q="source:kegg", fields="all")
    assert "kegg" in kegg["hits"][0].keys()
    assert kegg["hits"][0]["source"] == "kegg"


def test_query_by_source_do(geneset_client: MyGenesetInfo):
    do = geneset_client.query(q="source:do", fields="all")
    assert "do" in do["hits"][0].keys()
    assert do["hits"][0]["source"] == "do"


def test_query_by_source_reactome(geneset_client: MyGenesetInfo):
    reactome = geneset_client.query(q="source:reactome", fields="all")
    assert "reactome" in reactome["hits"][0].keys()
    assert reactome["hits"][0]["source"] == "reactome"


def test_query_by_source_smpdb(geneset_client: MyGenesetInfo):
    smpdb = geneset_client.query(q="source:smpdb", fields="all")
    assert "smpdb" in smpdb["hits"][0].keys()
    assert smpdb["hits"][0]["source"] == "smpdb"



class MyVariantClientMixin:
    """
    Adding some utility methods specific to MyVariant.info API.
    """

    def get_hgvs_from_vcf(self, input_vcf):
        """
        From the input VCF file (filename or file handle), return a generator
        of genomic based HGVS ids.

        :param input_vcf: input VCF file, can be a filename or a file handle
        :returns: a generator of genomic based HGVS ids. To get back a list instead,
                  using ``list(get_hgvs_from_vcf("your_vcf_file"))``

        .. NOTE:: This is a lightweight VCF parser to return valid genomic-based
                  HGVS ids from the *input_vcf* file. For more sophisticated VCF
                  parser, consider using `PyVCF <https://pypi.python.org/pypi/PyVCF>`_
                  module.
        """
        if isinstance(input_vcf, str):
            # if input_vcf is a string, open it as a file
            in_f = open(input_vcf)
        else:
            # otherwise it should be a file handle already
            in_f = input_vcf
        for row in in_f:
            if row[0] == "#":
                continue
            row = row.strip()
            if row:
                row = row.split("\t")
                if row[0].lower().startswith("chr"):
                    row[0] = row[0][3:]
                for alt in row[4].split(","):
                    yield self.format_hgvs(row[0], row[1], row[3], alt)

    @staticmethod
    def _normalized_vcf(chrom, pos, ref, alt):
        """
        If both ref/alt are > 1 base, and there are overlapping from the left,
        we need to trim off the overlapping bases.
        In the case that ref/alt is like this:  [CTTTT/CT] (with >1 overlapping bases from the left)
        ref/alt should be normalized as TTTT/T: [TC/TG --> C/G]
        and pos should be fixed as well.
        """
        for i in range(max(len(ref), len(alt))):
            _ref = ref[i] if i < len(ref) else None
            _alt = alt[i] if i < len(alt) else None
            if _ref is None or _alt is None or _ref != _alt:
                break

        # _ref/_alt cannot be both None, if so,
        # ref and alt are exactly the same,
        # something is wrong with this VCF record
        # assert not (_ref is None and _alt is None)
        if _ref is None and _alt is None:
            raise ValueError('"ref" and "alt" cannot be the same: {}'.format((chrom, pos, ref, alt)))

        _pos = int(pos)
        if _ref is None or _alt is None:
            # if either is None, del or ins types
            _pos = _pos + i - 1
            _ref = ref[i - 1 :]
            _alt = alt[i - 1 :]
        else:
            # both _ref/_alt are not None
            _pos = _pos + i
            _ref = ref[i:]
            _alt = alt[i:]

        return (chrom, _pos, _ref, _alt)

    def format_hgvs(self, chrom, pos, ref, alt):
        """
        get a valid hgvs name from VCF-style "chrom, pos, ref, alt" data.

        .. code-block:: python

           utils.variant.format_hgvs("1", 35366, "C", "T")
           utils.variant.format_hgvs("2", 17142, "G", "GA")
           utils.variant.format_hgvs("MT", 8270, "CACCCCCTCT", "C")
           utils.variant.format_hgvs("X", 107930849, "GGA", "C")
        """
        chrom = str(chrom)
        if chrom.lower().startswith("chr"):
            # trim off leading "chr" if any
            chrom = chrom[3:]
        if len(ref) == len(alt) == 1:
            # this is a SNP
            hgvs = "chr{0}:g.{1}{2}>{3}".format(chrom, pos, ref, alt)
        elif len(ref) > 1 and len(alt) == 1:
            # this is a deletion:
            if ref[0] == alt:
                start = int(pos) + 1
                end = int(pos) + len(ref) - 1
                if start == end:
                    hgvs = "chr{0}:g.{1}del".format(chrom, start)
                else:
                    hgvs = "chr{0}:g.{1}_{2}del".format(chrom, start, end)
            else:
                end = int(pos) + len(ref) - 1
                hgvs = "chr{0}:g.{1}_{2}delins{3}".format(chrom, pos, end, alt)
        elif len(ref) == 1 and len(alt) > 1:
            # this is an insertion
            if alt[0] == ref:
                hgvs = "chr{0}:g.{1}_{2}ins".format(chrom, pos, int(pos) + 1)
                ins_seq = alt[1:]
                hgvs += ins_seq
            else:
                hgvs = "chr{0}:g.{1}delins{2}".format(chrom, pos, alt)
        elif len(ref) > 1 and len(alt) > 1:
            if ref[0] == alt[0]:
                # if ref and alt overlap from the left, trim them first
                _chrom, _pos, _ref, _alt = self._normalized_vcf(chrom, pos, ref, alt)
                return self.format_hgvs(_chrom, _pos, _ref, _alt)
            else:
                end = int(pos) + len(ref) - 1
                hgvs = "chr{0}:g.{1}_{2}delins{3}".format(chrom, pos, end, alt)
        else:
            raise ValueError("Cannot convert {} into HGVS id.".format((chrom, pos, ref, alt)))
        return hgvs



"""
Tests for exercising the sychronous biothings_client for mygene
"""

import logging
import types

import pytest

import biothings_client
from biothings_client.client.definitions import MyGeneInfo
from biothings_client.utils.score import descore


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def test_metadata(gene_client: MyGeneInfo):
    meta = gene_client.metadata()
    assert "stats" in meta
    assert "total_genes" in meta["stats"]


def test_getgene(gene_client: MyGeneInfo):
    g = gene_client.getgene("1017")
    assert g["_id"] == "1017"
    assert g["symbol"] == "CDK2"


def test_getgene_with_fields(gene_client: MyGeneInfo):
    g = gene_client.getgene("1017", fields="name,symbol,refseq")
    assert "_id" in g
    assert "name" in g
    assert "symbol" in g
    assert "refseq" in g
    assert "summary" not in g


def test_curie_id_query(gene_client: MyGeneInfo):
    """
    Tests the annotation endpoint support for the biolink CURIE ID.

    If support is enabled then we should retrieve the exact same document for all the provided
    queries
    """
    curie_id_testing_collection = [
        ("1017", "entrezgene:1017", "NCBIgene:1017"),
        (1017, "entrezgene:1017", "ncbigene:1017"),
        ("1017", "entrezgene:1017", "NCBIGENE:1017"),
        ("1018", "ensembl.gene:ENSG00000250506", "ENSEMBL:ENSG00000250506"),
        (1018, "ensembl.gene:ENSG00000250506", "ensembl:ENSG00000250506"),
        ("5995", "uniprot.Swiss-Prot:P47804", "UniProtKB:P47804"),
        (5995, "uniprot.Swiss-Prot:P47804", "UNIPROTKB:P47804"),
        ("5995", "uniprot.Swiss-Prot:P47804", "uniprotkb:P47804"),
    ]

    results_aggregation = []
    for id_query, biothings_query, biolink_query in curie_id_testing_collection:
        id_query_result = gene_client.getgene(_id=id_query)
        biothings_term_query_result = gene_client.getgene(_id=biothings_query)
        biolink_term_query_result = gene_client.getgene(_id=biolink_query)
        results_aggregation.append(
            (
                id_query_result == biothings_term_query_result,
                id_query_result == biolink_term_query_result,
                biothings_term_query_result == biolink_term_query_result,
            )
        )

    results_validation = []
    failure_messages = []
    for result, test_query in zip(results_aggregation, curie_id_testing_collection):
        cumulative_result = all(result)
        if not cumulative_result:
            failure_messages.append(f"Query Failure: {test_query} | Results: {result}")
        results_validation.append(cumulative_result)

    assert all(results_validation), "\n".join(failure_messages)


def test_multiple_curie_id_query(gene_client: MyGeneInfo):
    """
    Tests the annotations endpoint support for the biolink CURIE ID.

    Batch query testing against the POST endpoint to verify that the CURIE ID can work with
    multiple

    If support is enabled then we should retrieve the exact same document for all the provided
    queries
    """
    curie_id_testing_collection = [
        ("1017", "entrezgene:1017", "NCBIgene:1017"),
        (1017, "entrezgene:1017", "ncbigene:1017"),
        ("1017", "entrezgene:1017", "NCBIGENE:1017"),
        ("1018", "ensembl.gene:ENSG00000250506", "ENSEMBL:ENSG00000250506"),
        (1018, "ensembl.gene:ENSG00000250506", "ensembl:ENSG00000250506"),
        ("5995", "uniprot.Swiss-Prot:P47804", "UniProtKB:P47804"),
        (5995, "uniprot.Swiss-Prot:P47804", "UNIPROTKB:P47804"),
        ("5995", "uniprot.Swiss-Prot:P47804", "uniprotkb:P47804"),
    ]

    results_aggregation = []
    for id_query, biothings_query, biolink_query in curie_id_testing_collection:
        base_result = gene_client.getgene(_id=id_query)

        batch_query = [id_query, biothings_query, biolink_query]
        query_results = gene_client.getgenes(ids=batch_query)
        assert len(query_results) == len(batch_query)

        batch_id_query = query_results[0]
        batch_biothings_query = query_results[1]
        batch_biolink_query = query_results[2]

        batch_id_query_return_value = batch_id_query.pop("query")
        assert batch_id_query_return_value == str(id_query)

        batch_biothings_query_return_value = batch_biothings_query.pop("query")
        assert batch_biothings_query_return_value == str(biothings_query)

        batch_biolink_query_return_value = batch_biolink_query.pop("query")
        assert batch_biolink_query_return_value == str(biolink_query)

        batch_result = (
            base_result == batch_id_query,
            base_result == batch_biothings_query,
            base_result == batch_biolink_query,
        )
        results_aggregation.append(batch_result)

    results_validation = []
    failure_messages = []
    for result, test_query in zip(results_aggregation, curie_id_testing_collection):
        cumulative_result = all(result)
        if not cumulative_result:
            failure_messages.append(f"Query Failure: {test_query} | Results: {result}")
        results_validation.append(cumulative_result)

    assert all(results_validation), "\n".join(failure_messages)


def test_getgene_with_fields_as_list(gene_client: MyGeneInfo):
    g1 = gene_client.getgene("1017", fields="name,symbol,refseq")
    g2 = gene_client.getgene("1017", fields=["name", "symbol", "refseq"])
    assert g1 == g2


def test_getgenes(gene_client: MyGeneInfo):
    g_li = gene_client.getgenes([1017, 1018, "ENSG00000148795"])
    assert len(g_li) == 3
    assert g_li[0]["_id"] == "1017"
    assert g_li[1]["_id"] == "1018"
    assert g_li[2]["_id"] == "1586"


def test_query(gene_client: MyGeneInfo):
    qres = gene_client.query("cdk2", size=5)
    assert "hits" in qres
    assert len(qres["hits"]) == 5


def test_query_with_fields_as_list(gene_client: MyGeneInfo):
    qres1 = gene_client.query("entrezgene:1017", fields="name,symbol,refseq")
    qres2 = gene_client.query("entrezgene:1017", fields=["name", "symbol", "refseq"])
    assert "hits" in qres1
    assert len(qres1["hits"]) == 1
    assert descore(qres1["hits"]) == descore(qres2["hits"])


def test_query_reporter(gene_client: MyGeneInfo):
    qres = gene_client.query("reporter:1000_at")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "5595"


def test_query_symbol(gene_client: MyGeneInfo):
    qres = gene_client.query("symbol:cdk2", species="mouse")
    assert "hits" in qres
    assert len(qres["hits"]) == 1
    assert qres["hits"][0]["_id"] == "12566"


def test_query_fetch_all(gene_client: MyGeneInfo):
    qres = gene_client.query("_exists_:pdb")
    total = qres["total"]

    qres = gene_client.query("_exists_:pdb", fields="pdb", fetch_all=True)
    assert isinstance(qres, types.GeneratorType)
    assert total == len(list(qres))


def test_querymany(gene_client: MyGeneInfo):
    qres = gene_client.querymany([1017, "695"], verbose=False)
    assert len(qres) == 2

    qres = gene_client.querymany("1017,695", verbose=False)
    assert len(qres) == 2


def test_querymany_with_scopes(gene_client: MyGeneInfo):
    qres = gene_client.querymany([1017, "695"], scopes="entrezgene", verbose=False)
    assert len(qres) == 2

    qres = gene_client.querymany([1017, "BTK"], scopes="entrezgene,symbol", verbose=False)
    assert len(qres) >= 2


def test_querymany_species(gene_client: MyGeneInfo):
    qres = gene_client.querymany([1017, "695"], scopes="entrezgene", species="human", verbose=False)
    assert len(qres) == 2

    qres = gene_client.findgenes([1017, "695"], scopes="entrezgene", species=9606, verbose=False)
    assert len(qres) == 2

    qres = gene_client.findgenes([1017, "CDK2"], scopes="entrezgene,symbol", species=9606, verbose=False)
    assert len(qres) == 2


def test_querymany_fields(gene_client: MyGeneInfo):
    qres1 = gene_client.findgenes(
        [1017, "CDK2"], scopes="entrezgene,symbol", fields=["uniprot", "unigene"], species=9606, verbose=False
    )
    assert len(qres1) == 2

    qres2 = gene_client.findgenes(
        "1017,CDK2", scopes="entrezgene,symbol".split(","), fields="uniprot,unigene", species=9606, verbose=False
    )
    assert len(qres2) == 2

    assert descore(qres1) == descore(qres2)


def test_querymany_notfound(gene_client: MyGeneInfo):
    qres = gene_client.findgenes([1017, "695", "NA_TEST"], scopes="entrezgene", species=9606)
    assert len(qres) == 3
    assert qres[2] == {"query": "NA_TEST", "notfound": True}


@pytest.mark.skipif(not biothings_client._PANDAS, reason="pandas library not installed")
def test_querymany_dataframe(gene_client: MyGeneInfo):
    from pandas import DataFrame

    query_list1 = [
        "1007_s_at",
        "1053_at",
        "117_at",
        "121_at",
        "1255_g_at",
        "1294_at",
        "1316_at",
        "1320_at",
        "1405_i_at",
        "1431_at",
    ]

    qres = gene_client.querymany(query_list1, scopes="reporter", as_dataframe=True)
    assert isinstance(qres, DataFrame)
    assert "name" in qres.columns
    assert set(query_list1) == set(qres.index)


def test_querymany_step(gene_client: MyGeneInfo):
    query_list1 = [
        "1007_s_at",
        "1053_at",
        "117_at",
        "121_at",
        "1255_g_at",
        "1294_at",
        "1316_at",
        "1320_at",
        "1405_i_at",
        "1431_at",
    ]
    qres1 = gene_client.querymany(query_list1, scopes="reporter")
    default_step = gene_client.step
    gene_client.step = 3
    qres2 = gene_client.querymany(query_list1, scopes="reporter")
    gene_client.step = default_step
    qres1 = descore(sorted(qres1, key=lambda doc: doc["_id"]))
    qres2 = descore(sorted(qres2, key=lambda doc: doc["_id"]))
    assert qres1 == qres2


def test_get_fields(gene_client: MyGeneInfo):
    fields = gene_client.get_fields()
    assert "uniprot" in fields.keys()
    assert "exons" in fields.keys()

    fields = gene_client.get_fields("kegg")
    assert "pathway.kegg" in fields.keys()



import io
import logging


def cache_request(fn):
    """
    Return a from_cache flag along with the request result
    The actual request is made in the given <fn> function
    """
    from_cache = False
    res = None
    logger = logging.getLogger("biothings.client")
    with io.StringIO() as buf:
        handler = logging.StreamHandler(buf)
        logger.addHandler(handler)
        try:
            res = fn()
            buf.flush()
            from_cache = "[ from cache ]" in buf.getvalue().strip()
        finally:
            logger.removeHandler(handler)
    return (from_cache, res)
